@@ -62,6 +62,14 @@ Currently only Linux uses this. */
 #define DEFAULT_VTABLE_THUNKS 1
 #endif
 
+#ifndef USE_GNULIBC_1
+#undef LIB_SPEC
+#define LIB_SPEC \
+  "%{shared: -lc} \
+   %{!shared: %{mieee-fp:-lieee} %{pthread:-lpthread} \
+	%{profile:-lc_p} %{!profile: -lc}}"
+#endif
+
 /* Output at beginning of assembler file.  */
 
 #undef ASM_FILE_START
