@@ -1,5 +1,5 @@
 /* Analyze RTL for C-Compiler
-   Copyright (C) 1987, 88, 9-5, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1987, 88, 9-6, 1998 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -22,6 +22,8 @@ Boston, MA 02111-1307, USA.  */
 #include "config.h"
 #include "rtl.h"
 
+extern int optimize;
+
 void note_stores ();
 int reg_set_p ();
 
@@ -1138,19 +1140,30 @@ dead_or_set_regno_p (insn, test_regno)
   int regno, endregno;
   rtx link;
 
-  /* See if there is a death note for something that includes TEST_REGNO.  */
-  for (link = REG_NOTES (insn); link; link = XEXP (link, 1))
+  /* REG_READ notes are not normally maintained after reload, so we
+     ignore them if the are invalid.  */
+  if (! reload_completed
+#ifdef PRESERVE_DEATH_INFO_REGNO_P
+      || PRESERVE_DEATH_INFO_REGNO_P (test_regno)
+#endif
+      )
     {
-      if (REG_NOTE_KIND (link) != REG_DEAD || GET_CODE (XEXP (link, 0)) != REG)
-	continue;
+      /* See if there is a death note for something that includes
+         TEST_REGNO.  */
+      for (link = REG_NOTES (insn); link; link = XEXP (link, 1))
+	{
+	  if (REG_NOTE_KIND (link) != REG_DEAD
+	      || GET_CODE (XEXP (link, 0)) != REG)
+	    continue;
 
-      regno = REGNO (XEXP (link, 0));
-      endregno = (regno >= FIRST_PSEUDO_REGISTER ? regno + 1
-		  : regno + HARD_REGNO_NREGS (regno,
-					      GET_MODE (XEXP (link, 0))));
+	  regno = REGNO (XEXP (link, 0));
+	  endregno = (regno >= FIRST_PSEUDO_REGISTER ? regno + 1
+		      : regno + HARD_REGNO_NREGS (regno,
+						  GET_MODE (XEXP (link, 0))));
 
-      if (test_regno >= regno && test_regno < endregno)
-	return 1;
+	  if (test_regno >= regno && test_regno < endregno)
+	    return 1;
+	}
     }
 
   if (GET_CODE (insn) == CALL_INSN
